#
# Control an X-10 CP-290
#
# $Id: cp-290.tcl,v 6.5 1997/02/17 16:31:21 loverso Exp loverso $
#
# Direct func: on=0x02 off=0x03
#
# 16*0xff	sync->
# 0x01		cmd_direct->
# 0x03		(dimlevel<<8)|off->
# 0x60		hc_A->
# 0x##		modules 9..16->
# 0x##		modules 1..8->
# 0x##		cksum(dimlevel..modules)->
# (22 bytes)
#
#		<-sync			6*0xff
#		<-status		0x00(rebooted)|0x01(ok)
#		<-hc_A|off		0x63
#		<-modules 9..16		0x##
#		<-modules 1..8		0x##
#		<-hc_A			0x60
#		<-cksum(hc..hc)		0x##
#					(12 bytes)

#
# CP-290 Protocol definitions
#
global HouseCode Cmd Func Sync

set Sync(1) "\377"
set Sync(2) $Sync(1)$Sync(1)
set Sync(4) $Sync(2)$Sync(2)
set Sync(6) $Sync(4)$Sync(2)
set Sync(16) $Sync(4)$Sync(4)$Sync(4)$Sync(4)

Set HouseCode Unknown -1
Set HouseCode A [expr 0x60]
Set HouseCode B [expr 0xe0]
Set HouseCode C [expr 0x20]
Set HouseCode D [expr 0xa0]
Set HouseCode E [expr 0x10]
Set HouseCode F [expr 0x90]
Set HouseCode G [expr 0x50]
Set HouseCode H [expr 0xd0]
Set HouseCode I [expr 0x70]
Set HouseCode J [expr 0xf0]
Set HouseCode K [expr 0x30]
Set HouseCode L [expr 0xb0]
Set HouseCode M [expr 0x00]
Set HouseCode N [expr 0x80]
Set HouseCode O [expr 0x40]
Set HouseCode P [expr 0xc0]

# Commands

Set Cmd SetHouseCode 0
Set Cmd Direct 1
	# Direct functions

	Set Func Unknown -1
	Set Func On 2
	Set Func Off 3
	# Dim
	# CP-290 uses 0==bright; I use 15==bright (and convert it on the fly)
	# I use a default dim of 8 (half)
	Set Func Dimmed 4
	set Dim 5
	Set Func Dim [expr $Dim+8*16]
	for {set _i 0} {$_i < 16} {incr _i} {
		Set Func Dim$_i [expr {$Dim + (15-$_i)*16}]
	}

Set Cmd SetClock 2
Set Cmd Download 3
Set Cmd GetClock 4
	# Days
	Set Days Sun [expr 0x40]
	Set Days Sat [expr 0x20]
	Set Days Fri [expr 0x10]
	Set Days Thu [expr 0x08]
	Set Days Wed [expr 0x04]
	Set Days Tue [expr 0x02]
	Set Days Mon [expr 0x01]
	Set Days ??? 0

Set Cmd GetTimer 5
Set Cmd GetData 6
Set Cmd Diag 7


# End protocol definitions

proc usleep {t} {
	after [expr $t * 1000]
}

#
# Forge connection to device
#
# Path can be either:
#	a UNIX pathname, i.e., "/dev/tty00"
#	{tcpport}@{hostname}, i.e., "5026@gizmo".
#

proc CP290-Connect {path {init ""}} {
	global cp290

	if [regexp -- {^([0-9]+)@(.+)} $path dummy port host] {
		if [catch {socket $host $port} res] {
			set cp290(error) $res
			return 0
		}
	} else {
		if [catch {set fid [open $path r+]} res] {
			set cp290(error) $res
			return 0
		}
	}
	set fid $res
	if [string compare "" $init] {
		if [catch [concat $init $fid] res] {
			close $fid
			set cp290(error) $res
			return 0
		}
		
	}
	fconfigure $fid -buffering none
	set cp290(io) $fid

	return 1
}

proc CP290-Disconnect {} {
	global cp290

	catch {close $cp290(io)}
	# Cause errors!
	catch {unset cp290}
}

#
# Input
#

proc CP290-GetByte {{to 5}} {
	global cp290
	return [GetByte $cp290(io) $to]
}

proc CP290-IFlush {} {
	global cp290
	return [IFlush $cp290(io)]
}

#
# Output
#

proc CP290-PutByte {val} {
	global cp290
	PutByte $cp290(io) $val
}

proc CP290-PutBytes {args} {
	global cp290
	eval [concat PutBytes $cp290(io) $args]
}

proc CP290-Puts {s} {
	global cp290
	puts -nonewline $cp290(io) $s
}

proc CP290-Sync {{duration 16}} {
	global cp290 Sync
	puts -nonewline $cp290(io) $Sync($duration)$Sync($duration)
##	usleep .1
}

#
# Protocol
#

proc CP290-Idle {} {
	CP290-Sync 16
}

proc CheckSum {args} {
	set sum 0
	foreach a $args {
		if [string match {0[^0*]} $a] {
			incr sum [string trimleft $a 0]
		} else {
			incr sum $a
		}
	}
	return [expr $sum % 256]
}
proc CheckSumU {units args} {
	return [eval CheckSum \
		[expr $units & 0xff] [expr ($units & 0xff00) >> 8] $args]
}

proc ConsumeACK {{to 5}} {
	global Sync cp290

	set ret 1

	set ready [Readable $cp290(io) [expr $to * 1000]]
	if {[string compare $ready $cp290(io)] != 0} {
		msg "Timed out waiting for Sync(6)"
		return 0
	}
	if {[read $cp290(io) 6] != $Sync(6)} {
		msg "Missing Sync(6)"
		return 0
	}
	set stat [CP290-GetByte]
	switch -exact $stat {
		0	{msg "CP-290 was reset (don't worry about this)";
			set cp290(reset) 1}
		1	{}
		default	{msg "unexpected ACK status"; set ret 0}
	}
	return $ret
}

#
# Direct Command
#

proc DirectCmdSend {fun hc units} {
	global Cmd Func HouseCode

	CP290-Sync 16
	CP290-PutByte $Cmd(Direct)
	CP290-PutByte $Func($fun)
	CP290-PutByte $HouseCode($hc)
	set uhi [expr {$units & 0xff}]
	set ulo [expr {($units & 0xff00) >> 8}]
	CP290-PutByte $uhi
	CP290-PutByte $ulo
	CP290-PutByte [CheckSumU $units $Func($fun) $HouseCode($hc)]

	if [ConsumeACK] {
		DirectStat "  Cmd Sent" $fun $hc $units
		return 1
	} else {
		DirectStat "Failed to Send" $fun $hc $units
		return 0
	}
}

proc DirectCmdResp {fun hc units} {
	global Func HouseCode

	if ![ConsumeACK] {
		return 0
	}
	set funhc [CP290-GetByte]
	set mh [CP290-GetByte]
	set ml [CP290-GetByte]
	set bhc [CP290-GetByte]
	set chk [CP290-GetByte]

	set fun $Func([expr {$funhc & 0xf}])
	set hc $HouseCode([expr {$funhc & 0xf0}])
	set units [expr {$ml*256+$mh}]
	# currently ignore base hc

	if {$chk != [CheckSumU $units $funhc $bhc]} {
		msg "Response: checksum did not match"
		return 0
	}

	DirectStat Successful $fun $hc $units
	return 1
}

proc DirectStat {op fun hc units} {
	puts -nonewline stderr "$op: $fun $hc units:"
	for {set i 1} {$i <= 16} {incr i} {
		if {$units & 1<<(16 - $i)} {
			puts -nonewline stderr [format " %d" $i]
		}
	}
	puts stderr ""
}

proc DirectCommand {fun hc units} {
	if [DirectCmdSend $fun $hc $units] {
		return [DirectCmdResp $fun $hc $units]
	} else {
		return 0
	}
}

#
# Request Clock/base housecode Command
#

proc GetClockSend {} {
	global Cmd

	CP290-Sync 16
	CP290-PutByte $Cmd(GetClock)
	return 1
}

proc GetClockResp {args} {
	global Cmd HouseCode Days

	if ![ConsumeACK 20] {
		return 0
	}
	set mins [CP290-GetByte]
	set hrs [CP290-GetByte]
	set day [CP290-GetByte]
	set bhc [CP290-GetByte]
	set chk [CP290-GetByte]

	if { $chk != [ CheckSum $mins $hrs $day $bhc] } {
		msg "Response checksum did not match"
		return 0
	}

	set hc $HouseCode([expr {$bhc & 0xf0}])
	set day $Days($day)

	puts stderr [format "GetClock: BHC=%s, Time=%s %02d:%02d\n" \
		$hc $day $hrs $mins]

	if {[llength $args] > 1} {
		# hr min
		upvar 2 [lindex $args 0] uhr \
			[lindex $args 1] umin
		set umin $mins
		set uhr $hrs
	}
	if {[llength $args] > 2} {
		# day
		upvar 2 [lindex $args 2] uday 
		set uday $day
	}
	if {[llength $args] > 3} {
		# hc
		upvar 2 [lindex $args 3] uhc
		set uhc $hc
	}
	return 1
}

# GetClock [hr min [day [bhc]]]
proc GetClock {args} {
	if [GetClockSend] {
		return [eval GetClockResp $args]
	} else {
		return 0
	}
}

proc SetClock {day hr min} {
	global Cmd Days

	CP290-Sync 16
	CP290-PutBytes $Cmd(SetClock) $min $hr $Days($day) \
		[CheckSum $min $hr $Days($day)]
	return [ConsumeACK]
}

proc SetBHC {hc} {
	global Cmd HouseCode

	CP290-Sync 16
	CP290-PutBytes $Cmd(SetHouseCode) $HouseCode($hc)
	return [ConsumeACK]
}

proc Retry {count args} {
##	CP290-Idle	;# hack
	while {$count >= 0} {
		# flush any waiting input
##		CP290-IFlush
		if [eval $args] {
			return 1
		}
		if {$count > 0} {
			msg "Failed talking with CP-290; will retry..."
			exec sleep 1
			CP290-IFlush
			CP290-Idle
		}
		incr count -1
	}
	return 0
}
#eof
